from functools import wraps


def space(f):
    @wraps(f)
    def wrapped(self, *args, **kwargs):
        print()
        return f(self, *args, **kwargs)

    return wrapped


class Formatter:
    """An API to control formatting of the command line interface"""

    def __init__(self, width):
        self.w = width

    # style formatting options {'name': ('string code', length of string code)}
    formats = {
        "PURPLE": ("\033[95m", 5),
        "BLUE": ("\033[94m", 5),
        "GREEN": ("\033[92m", 5),
        "YELLOW": ("\033[93m", 5),
        "RED": ("\033[91m", 5),
        "BOLD": ("\033[1m", 4),
        "END": ("\033[0m", 4),
    }

    def center(self, msg, styles=[], fill=' ', in_w=0, in_b=''):
        """Style text and center across width and optional inner container"""

        # apply styles to message and inner container border
        fmt_start = ''
        fmt_chars = 0
        for style in styles:
            if style not in self.formats:
                raise NotImplementedError(f'Style "{style}" not implemented.')
            fmt_start += self.formats[style][0]
            fmt_chars += self.formats[style][1]
        msg = fmt_start + msg + self.formats['END'][0]
        fmt_chars += self.formats['END'][1]

        # margin refers to combined space on each side of inner container
        # note that border size is removed from inner width, not margin
        margin = self.w - in_w if in_w else 0
        lmargin = margin // 2

        # print left margin
        print(' ' * lmargin, end='')

        # calculate width for '.center' method inside of 'print' function
        #  + self.w (starting total width)
        #  - subtract margin
        #  - subtract size of inner container border
        #  + add number of formatting characters in msg that don't print
        # Each formatting code adds a certain number of characters to msg that
        # the print function considers when centering the output. If we do not
        # adjust for this, then the actual width of printed characters will
        # be less than the desired width (by exactly the number of fmt_chars)
        width = self.w - margin - 2 * len(in_b) + fmt_chars
        print(in_b + msg.center(width, fill) + in_b)

    def prompt(self, msg):
        """Prompt the user, applying consistent formatting"""
        msg = ">> " + msg + " >>"
        n = len(msg)
        msg, off = self.apply(msg, ["RED"])
        print(msg.rjust((self.w // 2) + (n // 2) + off), end="")
        return input(" ")

    def apply(self, text, options=["BOLD"]):
        """Return string with formatting applied and number of format chars"""
        fmt_start = ''
        fmt_chars = 0
        for format in options:
            if format not in self.formats:
                raise NotImplementedError(f'Style "{format}" not implemented.')
            fmt_start += self.formats[format][0]
            fmt_chars += self.formats[format][1]
        fmt_chars += self.formats['END'][1]
        return fmt_start + text + self.formats["END"][0], fmt_chars

    def print_row(self, text, inner_w, row, formatting=False):
        """Print a row of the board accounting for the formatting applied"""
        margin = (self.w - inner_w) // 2
        print(" " * margin + "|", end="")
        if formatting:
            padding = (inner_w - 2 - (row + 1) * 2) // 2
            print(" " * padding, end="")
            print(text, end="")
            print(" " * padding, end="")
        else:
            print(text.center(inner_w - 2), end="")
        print("|")



from .formatter import space


class Board:
    """Represent the board of the IQ Test game"""

    def __init__(self, formatter, rows):
        self.f = formatter
        self.rows = rows
        self.holes = self.rows * (self.rows + 1) // 2
        self.pegs = [chr(i) for i in range(97, 97 + self.holes)]
        self.board = self.initiate()

    def initiate(self):
        """Create triangular array with peg (letter) in each hole (element)"""
        board = []
        for i in range(self.rows):
            start = i * (i + 1) // 2
            row = [x for x in self.pegs[start: start + i + 1]]
            board.append(row)
        return board

    def pegs_left(self):
        """Tally up the number of remaining pegs on the board"""
        count = 0
        for row in self.board:
            for val in row:
                count += val is not None
        return count

    def locate_peg(self, peg):
        """Search board for peg and return location or None if not found"""
        for i in range(self.rows):
            for j in range(i + 1):
                if self.board[i][j] == peg:
                    return (i, j)
        return None

    @space
    def show(self, highlight=set(), color="RED"):
        """Print board to command line"""
        w = 30
        print(" IQ Tester Board ".center(w - 2, "-").center(self.f.w))
        self.f.center("", in_w=w, in_b='|')

        # iterate over each row of the board
        for i in range(self.rows):
            disp = ""
            fmt_chars = 0
            # iterate over each peg and add it to display string
            for j in range(i + 1):
                val = self.board[i][j]
                # there is a peg in the position
                if val:
                    # check if it should be highlighted
                    if (i, j) in highlight:
                        val, inc = self.f.apply(val, ["BOLD", color])
                        fmt_chars += inc
                    disp += val + " "
                # empty hole
                else:
                    disp += ". "
            # print row
            self.f.print_row(disp, w, i, fmt_chars > 0)

        # finish boarder
        self.f.center("", in_w=w, in_b='|')
        print(("-" * (w - 2)).center(self.f.w))

    def remove(self, peg):
        """Remove peg from the board"""
        for i in range(self.rows):
            for j in range(i + 1):
                if self.board[i][j] == peg:
                    self.board[i][j] = None
                    return True
        return False

    def get_moves(self):
        """Return a list of the possible moves on the board"""
        moves = {}
        # for each hole, check for possible moves, which require a peg in hole,
        # a peg in neighbor, and the landing place to be empty
        for i in range(self.rows):
            for j in range(i + 1):
                if self.board[i][j]:
                    moves[(i, j)] = []
                    # check down-left
                    if (
                        i < self.rows - 2
                        and self.board[i + 1][j]
                        and self.board[i + 2][j] is None
                    ):
                        moves[(i, j)].append(((i + 1, j), (i + 2, j)))
                    # check down-right
                    if (
                        i < self.rows - 2
                        and self.board[i + 1][j + 1]
                        and self.board[i + 2][j + 2] is None
                    ):
                        moves[(i, j)].append(((i + 1, j + 1), (i + 2, j + 2)))
                    # check up-left
                    if (
                        i > 1
                        and j > 1
                        and self.board[i - 1][j - 1]
                        and self.board[i - 2][j - 2] is None
                    ):
                        moves[(i, j)].append(((i - 1, j - 1), (i - 2, j - 2)))
                    # check up-right
                    if (
                        i > 1
                        and j <= i - 2
                        and self.board[i - 1][j]
                        and self.board[i - 2][j] is None
                    ):
                        moves[(i, j)].append(((i - 1, j), (i - 2, j)))
                    # check left
                    if (
                        j > 1
                        and self.board[i][j - 1]
                        and self.board[i][j - 2] is None
                    ):
                        moves[(i, j)].append(((i, j - 1), (i, j - 2)))
                    # check right
                    if (
                        j < i - 1
                        and self.board[i][j + 1]
                        and self.board[i][j + 2] is None
                    ):
                        moves[(i, j)].append(((i, j + 1), (i, j + 2)))
                    # if no moves, remove dictionary key
                    if moves[(i, j)] == []:
                        del moves[(i, j)]
        return moves
